"""
    create_manifest_table(conn::MySQL.Connection)

Create a 'manifest' table in a MySQL database if it does not already exist.

This function creates a table named 'manifest' with columns for id, method_name, arguments_string, arguments_dict, and result_table_name. The id column is an auto-incrementing primary key.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> create_table(conn)

In this example, `create_manifest_table` is called with a MySQL connection `conn`. This will create the 'manifest' table in the connected database if it doesn't already exist.
"""
function create_manifest_table(conn::MySQL.Connection)
    # Prepare query
    query = """
        CREATE TABLE IF NOT EXISTS manifest (
            id INT AUTO_INCREMENT PRIMARY KEY,
            method_name VARCHAR(255) NOT NULL,
            arguments_string VARCHAR(255) NOT NULL,
            arguments_dict VARCHAR(64) NOT NULL,
            result_table_name VARCHAR(255) NOT NULL
        );
    """

    # Create the 'manifest' table if it doesn't exist
    query_connection(conn, query)
end

"""
create_results_database(service::OrchestratorRegistry.Service)

Ensures the `FACT_results` database exists for the specified service.

This function connects to the database server using the provided `Service` instance, checks for the 
existence of the `FACT_results` database, and creates it if it does not already exist.

Arguments:
  - `service::OrchestratorRegistry.Service`: The `Service` instance containing the host and port 
    details required to connect to the database server.

Behavior:
  - Connects to the database server using the service's host and port.
  - Executes a `CREATE DATABASE IF NOT EXISTS` query to ensure the `FACT_results` database exists.
  - Closes the connection to the database server after executing the query.

Raises:
  - `Error`: If the connection to the database server fails or the query execution encounters an error.

Example:

```julia
# Define a service instance
service = Service("123", "auth-service", "http://localhost", 3306)

# Ensure the database exists
create_database(service)
```
"""
function create_results_database(service::OrchestratorRegistry.Service)
    conn = establish_connection(service.host, "root", "devops", ""; port=service.port)

    # Prepare query
    query = "CREATE DATABASE IF NOT EXISTS FACT_results;"

    # Create the databse if it doesn't exist
    DBInterface.execute(conn, query)

    # Close connection
    DBInterface.close!(conn)
end  

"""
get_results_connection(service::OrchestratorRegistry.Service)

Establishes and returns a database connection for the specified service.

This function ensures that the database associated with the given `Service` instance is ready for use.
It performs the following steps:
  1. Creates the database if it does not already exist.
  2. Establishes a connection to the database using the service's host and port.
  3. Ensures the `manifest` table is present in the database, creating it if necessary.

Arguments:
  - `service::OrchestratorRegistry.Service`: The `Service` instance containing the details 
    (host and port) required to connect to the database.

Returns:
  - `conn`: The database connection object established for the specified service.

Raises:
  - `Error`: If the database connection cannot be established or if any database operation fails.

Behavior:
  - Uses the host and port of the service to connect.
  - Assumes the database credentials are `root` and `devops`.
  - The database name used is `FACT_results`.

Example:

```julia
# Define a service instance
service = Service("123", "auth-service", "http://localhost", 3306)

# Establish a database connection
conn = get_connection(service)

# Perform operations using the connection
execute_query(conn, "SELECT * FROM manifest")
```
"""
function get_results_connection(service::OrchestratorRegistry.Service)
    # Create the databse if it does not exist
    create_results_database(service)

    # Connect to the database
    conn = establish_connection(service.host, "root", "devops", "FACT_results"; port=service.port)

    # Create the 'manifest' table if it does not exist
    create_manifest_table(conn)

    return conn
end

"""
get_results_connection(service_dict::Dict) -> Connection

Establishes a database connection using the provided service configuration.

This function creates a `Service` object from the given `service_dict` and uses it to establish a database connection. The connection is configured based on the `host` and `port` specified in the dictionary.

# Arguments
- `service_dict::Dict`: A dictionary containing the following keys:
  - `"host"`: The hostname or IP address of the database service.
  - `"port"`: The port number for the database service.

# Returns
- `Connection`: A connection object to the database service.

# Behavior
- Constructs a `Service` object using the provided host and port from `service_dict`.
- Calls `get_results_connection` with the constructed `Service` object to establish the database connection.

# Example
Define the service dictionary and establish a connection:

```julia
service_dict = Dict(
    "host" => "127.0.0.1",
    "port" => 3306
)

connection = get_results_connection(service_dict)
println(connection)
```

In this example, a database connection is established using the host `127.0.0.1` and port `3306`.
"""
function get_results_connection(service_dict::Dict)
    service = OrchestratorRegistry.Service(
        "id",
        "name",
        service_dict["host"],
        service_dict["port"]
    )
    return get_results_connection(service)
end

"""
close_connection(conn::MySQL.Connection)

Close the connection to a MySQL database.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to be closed.

This function closes the MySQL database connection specified by `conn` using the `DBInterface.close!` function.

"""
function close_connection(conn::MySQL.Connection)
    DBInterface.close!(conn)
end

"""
    query_connection(conn::MySQL.Connection, query::String) -> DataFrame

Execute a SQL query on a given MySQL database connection and return the results in a DataFrame.

This function takes a MySQL connection and a SQL query string, executes the query, and returns the results in a DataFrame. Each row of the query result is converted into a row in the DataFrame. All column types in the resulting DataFrame are set to `String`.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `query::String`: The SQL query string to be executed.

# Returns
- `DataFrame`: A DataFrame containing the results of the query.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> query = "SELECT * FROM my_table"
> df = query_connection(conn, query)

In this example, `query_connection` is used to execute a SQL query on a MySQL database and the results are stored in a DataFrame `df`.
"""
function query_connection(conn::MySQL.Connection, query::String)
    cursor = DBInterface.execute(conn, query)
    cursor.types = [String for _ in cursor.types]
    rows = Tables.rows(cursor)
    # Initialize an empty DataFrame
    df = DataFrame()
    # Process rows
    for row in rows
        # If df is empty, initialize columns
        if ncol(df) == 0
            for col_name in cursor.names
                df[!, Symbol(col_name)] = String[]
            end
        end
        # Process the row
        row_data = [string(getproperty(row, Symbol(col_name))) for col_name in cursor.names]
        push!(df, row_data)
    end
    return df
end

"""
    establish_connection(host, user, password, dbname; port=3306) -> MySQL.Connection

Establish a connection to a MySQL/MariaDB database.

# Arguments
- `host::String`: Database host
- `user::String`: Username
- `password::String`: Password
- `dbname::String`: Database name
- `port::Int=3306`: Port number (default is 3306)

# Returns
- `MySQL.Connection`: A connection object to the database.
"""
function establish_connection(host::String, user::String, password::String, dbname::String; port::Int=3306)
    conn = MySQL.Connection(host, user, password, dbname, port, "nothing")
    size_query = "SET GLOBAL max_allowed_packet=1073741824;"  # Sets max_allowed_packet to 1GB
    timeout_query = "SET SESSION wait_timeout = 31536000;"     # Sets wait_timeout to 1 year
    maxtime_query = "SET SESSION MAX_STATEMENT_TIME=604800000;"  # Sets max_execution_time to 1 week
     try
         # Execute session variable adjustments
         DBInterface.execute(conn, size_query)
     catch ex
         @error "Failed to set session variables for query $size_query with expection " exception=(ex, catch_backtrace())
         # Handle error, possibly rethrow or clean up
     end
     try
         # Execute session variable adjustments
         DBInterface.execute(conn, timeout_query)
     catch ex
         @error "Failed to set session variables for query $timeout_query with expection " exception=(ex, catch_backtrace())
         # Handle error, possibly rethrow or clean up
     end
     try
         # Execute session variable adjustments
         DBInterface.execute(conn, maxtime_query)
     catch ex
         @error "Failed to set session variables for query $maxtime_query with expection " exception=(ex, catch_backtrace())
         # Handle error, possibly rethrow or clean up
     end
    return conn
end

"""
    SQLTYPES::Dict{Type, String}

A dictionary mapping from Julia data types to SQL data types, used for generating SQL table schemas that correspond to the structure of a Julia DataFrame. This mapping ensures that data stored in a DataFrame can be appropriately represented in a SQL database.

# Supported Mappings
- `Int8` to `TINYINT`
- `Int16` to `SMALLINT`
- `Int32` to `INTEGER`
- `Int64` to `BIGINT`
- `UInt8` to `TINYINT UNSIGNED`
- `UInt16` to `SMALLINT UNSIGNED`
- `UInt32` to `INTEGER UNSIGNED`
- `UInt64` to `BIGINT UNSIGNED`
- `Float32` to `FLOAT`
- `Float64` to `DOUBLE`
- `Bool` to `BOOL`
- `Vector{UInt8}` to `BLOB` (for storing binary data)
- `String` to `VARCHAR(255)` (default maximum length for strings)
- `Date` to `DATE`
- `Time` to `TIME`
- `DateTime` to `DATETIME`

The dictionary is used by the `create_table_from_dataframe` function to automatically determine the SQL column types based on the DataFrame's column types.
"""
const SQLTYPES = Dict{Type, String}(
    Int8 => "TINYINT",
    Int16 => "SMALLINT",
    Int32 => "INTEGER",
    Int64 => "BIGINT",
    UInt8 => "TINYINT UNSIGNED",
    UInt16 => "SMALLINT UNSIGNED",
    UInt32 => "INTEGER UNSIGNED",
    UInt64 => "BIGINT UNSIGNED",
    Float32 => "FLOAT",
    Float64 => "DOUBLE",
    Bool => "BOOL",
    Vector{UInt8} => "BLOB",
    String => "VARCHAR(255)",
    Date => "DATE",
    Time => "TIME",
    DateTime => "DATETIME"
)

"""
    SQL_RANGES

A constant dictionary mapping Julia data types to tuples representing the minimum and maximum values allowed for each type, intended to align with the range limits of corresponding SQL data types.

This mapping ensures data being prepared for SQL operations does not exceed the storage capabilities of the target database's data types, helping prevent runtime errors due to type overflows or underflows.

# Contents
The dictionary contains mappings for the following Julia types:
- `Int8`, `Int16`, `Int32`, `Int64`: Signed integer types with corresponding SQL range limits.
- `UInt8`, `UInt16`, `UInt32`, `UInt64`: Unsigned integer types with corresponding SQL range limits.
- `Float32`, `Float64`: Floating-point types with approximate range limits to accommodate the range of floating-point numbers in SQL.

# Usage Example
To clamp a value to the allowed range for its type before inserting it into a database:
> value = clamp(value, SQL_RANGES[typeof(value)][1], SQL_RANGES[typeof(value)][2])

# Notes
- The ranges for `Float32` and `Float64` are approximations and may vary depending on specific database implementations or configurations.
- This dictionary should be used as a guide for data preprocessing steps, ensuring compatibility with SQL database types and preventing insertion errors.

"""
const SQL_RANGES = Dict{Type, Tuple{Number, Number}}(
    Int8 => (-128, 127),
    Int16 => (-32768, 32767),
    Int32 => (-2147483648, 2147483647),
    Int64 => (-9223372036854775808, 9223372036854775807),
    UInt8 => (0, 255),
    UInt16 => (0, 65535),
    UInt32 => (0, 4294967295),
    UInt64 => (0, 18446744073709551615),
    Float32 => (-3.4e38, 3.4e38),  # Approximation
    Float64 => (-1.7e308, 1.7e308) # Approximation
)


"""
    create_table_from_dataframe(conn::MySQL.Connection, df::DataFrame, table_name::String)

Creates a new table in a MySQL or MariaDB database that matches the schema of a provided Julia DataFrame. This function dynamically constructs a `CREATE TABLE` SQL statement, mapping Julia types to SQL types according to the `SQLTYPES` dictionary, with special handling for columns intended to store geometrical data as `LONGTEXT`.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL or MariaDB database.
- `df::DataFrame`: The DataFrame whose schema is used to define the structure of the new table.
- `table_name::String`: The name of the new table to be created. If a table with this name already exists, the operation will fail unless the existing table is dropped beforehand.

# Special Handling
Columns named `shape_obj` or `shape` (case-insensitive) are treated specially and are mapped to the `LONGTEXT` SQL type, regardless of their original Julia type, to accommodate potentially large geometrical data.

# Usage Example
> conn = MySQL.connect("hostname", "username", "password", "database_name")
> df = DataFrame(region=["North", "South"], geo_id=["N001", "S001"], shape_obj=["POLYGON((...))", "POLYGON((...))"], area=[123.4, 567.8])
> create_table_from_dataframe(conn, df, "geographical_features")

This creates a table named 'geographical_features' in the specified database, with columns mapped from the DataFrame's schema and geometrical data columns set as `LONGTEXT`.

# Error Handling
If the DataFrame contains a column with an unsupported Julia type (not found in `SQLTYPES`), an `ArgumentError` is thrown.
"""
function create_table_from_dataframe(conn::MySQL.Connection, df::DataFrame, table_name::String)

    # Start building the CREATE TABLE statement
    create_stmt = "CREATE TABLE $(table_name) ("

    # Iterate over the columns of the DataFrame to define the fields
    for (name, type) in zip(names(df), eltype.(eachcol(df)))
        # Handle special cases for geometrical data first
        if lowercase(name) in ["shape_obj", "shape"]
            sql_type = "LONGTEXT"
        else
            # Find the SQL type from the dictionary, or throw an error if not found
            sql_type = get(SQLTYPES, type, nothing)
            if isnothing(sql_type)
                throw(ArgumentError("Unsupported column type: $type for column $name"))
            end
        end
        create_stmt *= "`$(name)` $sql_type, "
    end

    # Remove the trailing comma and space, add closing parenthesis
    create_stmt = create_stmt[1:end-2] * ");"

    # Execute the CREATE TABLE statement
    DBInterface.execute(conn, create_stmt)
end