"""
    get_table(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict) -> DataFrame

Retrieve data from a specified table in a MySQL database based on the given method name, arguments string, and arguments dictionary.

This function first attempts to find the table name in a 'manifest' table using the provided method name, arguments string, and arguments dictionary. If the table name is found, it then retrieves the data from this table and returns it as a DataFrame. If no corresponding table name is found, an empty DataFrame is returned.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `method_name::String`: The name of the method used to identify the table in the 'manifest'.
- `arguments_string::String`: The arguments string used to identify the table in the 'manifest'.
- `arguments::Dict`: A dictionary of arguments used to identify the table in the 'manifest'.

# Returns
- `DataFrame`: A DataFrame containing the data from the identified table, or an empty DataFrame if no table is found.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> method_name = "my_method"
> arguments_string = "arg1=val1,arg2=val2"
> arguments = Dict("arg1" => "val1", "arg2" => "val2")
> df = get_table(conn, method_name, arguments_string, arguments)

In this example, `get_table` is called with a MySQL connection `conn` and specific method details. It retrieves data from the identified table in the database and returns it as a DataFrame.
"""
function get_table(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict)
    table_str = find_table_name_in_manifest(conn, method_name, arguments_string, arguments)

    df = DataFrame()
    if table_str == ""
        return df
    else    
        df = query_connection(conn, "SELECT * FROM $table_str;")
    end
end

"""
    write_table(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict, results::DataFrame)

Write data from a DataFrame to a MySQL table based on the specified method name and arguments. This function first attempts to find a corresponding table name in the manifest. If the table does not exist, a new table name is generated, and the DataFrame is written to the new table in the MySQL database. If a table with the specified method name and arguments already exists, the function raises an error to prevent overwriting existing data.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `method_name::String`: The name of the method associated with the data. Used to identify or generate the table name.
- `arguments_string::String`: A string representation of the arguments associated with the data. Used in conjunction with `method_name` to identify or generate the table name.
- `arguments::Dict`: A dictionary of arguments associated with the data. This is used for more detailed tracking or identification but does not directly affect the table name.
- `results::DataFrame`: The DataFrame containing the data to be written to the MySQL database.

# Returns
- This function does not return a value. It either successfully writes the data to the database or raises an error if the table already exists.

# Errors
- Raises an error if a table associated with the provided `method_name` and `arguments_string` already exists in the manifest, indicating that the data might be overwritten.

# Example
> conn = MySQL.connect("host", "user", "password", "dbname")
> method_name = "my_data_method"
> arguments_string = "param1=value1,param2=value2"
> arguments = Dict("param1" => "value1", "param2" => "value2")
> results = DataFrame(A = 1:3, B = ["x", "y", "z"])

In this example, the `write_table` function attempts to write the contents of `results` DataFrame to a MySQL table. If a table corresponding to `method_name` and `arguments_string` does not exist, it will be created and populated with the data from `results`. If such a table already exists, an error will be raised.
"""
function write_table(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict, results::DataFrame)
    table_str = find_table_name_in_manifest(conn, method_name, arguments_string, arguments)
    if table_str == ""
        table_name = write_table_name_in_manifest(conn, method_name, arguments_string, arguments)

        create_table_from_dataframe(conn, results, table_name)
        try
            #MySQL.load(results, conn, table_name)
            custom_load(conn, table_name, results)
        catch e
            println("An error occurred: ", e)
        end
    else
        error("Table already exists in FACT_results.manifest: $table_str")
    end
end

"""
    does_table_exist_for_method(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict) -> Bool

Check if a specific table exists in the database based on a method name, an argument string, and additional arguments. This function queries a manifest (or similar mapping) that links method names and their arguments to database table names.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection.
- `method_name::String`: The name of the method to check. This is a key or identifier used to search the manifest.
- `arguments_string::String`: A string representation of arguments or conditions. Its format and usage can vary depending on the manifest's structure.
- `arguments::Dict`: A dictionary of arguments providing additional search criteria or parameters for the manifest query.

# Returns
`Bool`: Returns `true` if the table exists for the specified method and arguments, based on the manifest's mapping. Returns `false` if no such table is found.

# Example
To check if a table exists for a method 'getEmployeeData' with arguments indicating a permanent type in the engineering department, you would call the function as follows:
- Initialize your MySQL connection as `conn`.
- Set `method_name` to "getEmployeeData".
- Define `arguments_string` as "type=permanent".
- Create a dictionary `arguments` with `{"department" => "engineering"}`.
- Call `does_table_exist_for_method(conn, method_name, arguments_string, arguments)` to check for the table's existence.

# Notes
This function relies on the implementation of `find_table_name_in_manifest`, which must query a manifest or similar structure to find the corresponding table name based on the provided method name and arguments. The manifest structure and query logic are assumed to be predefined.
"""
function does_table_exist_for_method(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict)
    table_str = find_table_name_in_manifest(conn, method_name, arguments_string, arguments)
    return table_str != ""
end

"""
    overwrite_table(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict, results::DataFrame)

Overwrite a MySQL table with new data from a DataFrame. If the table specified by the method name and arguments exists, it is dropped and recreated with the new data. If the table does not exist, it is created.

This function first checks if a table name corresponding to the given method name and arguments exists in the manifest. 
If the table exists, it is dropped (along with its associated record in 'manifest'), and then a new table with the same name is created. The data from the provided DataFrame is then written into this new table. 
If the table does not exist, it is simply created and the data is written to it.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `method_name::String`: The name of the method, used as part of the table identification.
- `arguments_string::String`: A string representation of the arguments, used as part of the table identification.
- `arguments::Dict`: A dictionary of arguments, used as part of the table identification.
- `results::DataFrame`: The DataFrame containing data to be written to the database.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "FACT_results")
> method_name = "update_statistics"
> arguments_string = "param1=val1,param2=val2"
> arguments = Dict("param1" => "val1", "param2" => "val2")
> results = DataFrame(col1 = [1, 2, 3], col2 = ["A", "B", "C"])
> overwrite_table(conn, method_name, arguments_string, arguments, results)

In this example, `overwrite_table` is called with a MySQL connection, method name, arguments, and a DataFrame. If a table matching the method name and arguments exists, it is overwritten with the new data from the DataFrame.
"""
function overwrite_table(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict, results::DataFrame)
    table_str = find_table_name_in_manifest(conn, method_name, arguments_string, arguments)
    
    # delete already existing data
    if table_str != ""
        query = "DROP TABLE $table_str;"
        query_connection(conn, query)
        query2 = "DELETE FROM manifest WHERE result_table_name = '$table_str';"
        query_connection(conn, query2)
    end

    # write new data
    write_table(conn, method_name, arguments_string, arguments, results)
end

"""
    find_table_name_in_manifest(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict) -> String

Search for a table name in the 'manifest' table of a MySQL database that matches the given method name, arguments string, and arguments dictionary.

This function queries the 'manifest' table for an entry that matches the specified method name, arguments string, and a JSON-serialized version of the arguments dictionary. If a matching entry is found, the function returns the associated table name. If no match is found, an empty string is returned.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `method_name::String`: The name of the method to match in the 'manifest' table.
- `arguments_string::String`: The string representation of the arguments to match in the 'manifest' table.
- `arguments::Dict`: The dictionary of arguments to match in the 'manifest' table.

# Returns
- `String`: The name of the matching table in the 'manifest' table, or an empty string if no match is found.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> method_name = "fetch_data"
> arguments_string = "param1=value1,param2=value2"
> arguments = Dict("param1" => "value1", "param2" => "value2")
> table_name = find_table_name_in_manifest(conn, method_name, arguments_string, arguments)

In this example, `find_table_name_in_manifest` is used to search for a table in the 'manifest' table that matches the given method name and arguments. It returns the name of the found table or an empty string if no match is found.
"""
function find_table_name_in_manifest(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict)
    # Serialize the arguments dictionary to a JSON string
    arguments_dict = bytes2hex(sha256(JSON.json(arguments)))
 
    # Prepare and execute a query to check if the entry exists in the manifest table
    query = """
    SELECT result_table_name
    FROM manifest
    WHERE method_name = '$method_name' AND arguments_string = '$arguments_string' AND arguments_dict = '$arguments_dict';
    """
    result = query_connection(conn, query)

    # Check if a result was found
    if !isempty(result)
        return result[1, :result_table_name]
    else
        return "" 
    end
end

"""
    check_if_table_name_in_manifest(conn::MySQL.Connection, table_name::String) -> DataFrame

Check if a given table name exists in the 'manifest' table of a MySQL database.

This function queries the 'manifest' table to determine if the specified table name exists within it. The query checks for the presence of `table_name` in the `result_table_name` column of the 'manifest' table. The function returns the query result as a DataFrame, which will be non-empty if the table name exists.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `table_name::String`: The table name to search for in the 'manifest' table.

# Returns
- `DataFrame`: A DataFrame containing the query result. It will be non-empty if the table name exists in the 'manifest' table.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> table_name = "my_table"
> result_df = check_if_table_name_in_manifest(conn, table_name)

In this example, `check_if_table_name_in_manifest` is used to check if 'my_table' exists in the 'manifest' table of the database. The result of the query is returned as a DataFrame.
"""
function check_if_table_name_in_manifest(conn::MySQL.Connection, table_name::String)
    # Prepare and execute a query to check if the entry exists in the manifest table
    query = """
    SELECT result_table_name
    FROM manifest
    WHERE result_table_name = '$table_name';
    """
    result = query_connection(conn, query)
    
    return result
end

"""
    write_table_name_in_manifest(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict) -> String

Insert a new entry into the 'manifest' table with a unique, randomly generated table name, along with the provided method name, arguments string, and arguments dictionary. The function returns the generated table name.

This function first serializes the provided arguments dictionary into a JSON string. It then generates a random table name and checks if this name already exists in the 'manifest' table. If the name exists, it generates a new name. This process repeats until a unique name is found. Once a unique table name is generated, the function inserts an entry into the 'manifest' table with the method name, arguments string, arguments dictionary, and the generated table name.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL database.
- `method_name::String`: The name of the method to be recorded in the 'manifest' table.
- `arguments_string::String`: The string representation of the arguments to be recorded in the 'manifest' table.
- `arguments::Dict`: The dictionary of arguments to be serialized and recorded in the 'manifest' table.

# Returns
- `String`: The randomly generated unique table name.

# Example
> conn = MySQL.connect("localhost", "username", "password", db = "dbname")
> method_name = "fetch_data"
> arguments_string = "param1=value1,param2=value2"
> arguments = Dict("param1" => "value1", "param2" => "value2")
> table_name = write_table_name_in_manifest(conn, method_name, arguments_string, arguments)

In this example, `write_table_name_in_manifest` is called with a MySQL connection, method name, arguments string, and a dictionary of arguments. It inserts a new entry into the 'manifest' table and returns the generated table name.
"""
function write_table_name_in_manifest(conn::MySQL.Connection, method_name::String, arguments_string::String, arguments::Dict)
    # Serialize the arguments dictionary to a JSON string
    arguments_dict = bytes2hex(sha256(JSON.json(arguments)))

    # Generate a random table name
    result_table_name = generate_random_table_name(10)

    # Make sure is new name
    while !isempty(check_if_table_name_in_manifest(conn, result_table_name))
        result_table_name = generate_random_table_name(10)
    end

    # Prepare and execute a query to insert the entry into the manifest table
    query = """
    INSERT INTO manifest (method_name, arguments_string, arguments_dict, result_table_name)
    VALUES ('$method_name', '$arguments_string', '$arguments_dict', '$result_table_name');
    """
    query_connection(conn, query)

    return result_table_name
end

"""
    generate_random_table_name(length::Int) -> String

Generate a random table name of the specified length, adhering to MySQL naming conventions.

The function creates a random table name consisting of letters (both uppercase and lowercase), digits, and underscores. The first character of the table name is ensured to be a letter or an underscore, as per MySQL table naming rules.

# Arguments
- `length::Int`: The desired length of the generated table name.

# Returns
- `String`: A randomly generated table name that complies with MySQL naming rules.

# Example
> table_name = generate_random_table_name(10)

This example generates a random table name with 10 characters. The resulting table name is a valid identifier for a MySQL table.
"""
function generate_random_table_name(length::Int)
    # Define valid characters for a MySQL table name
    valid_chars = ['A':'Z'; 'a':'z'; '0':'9'; '_']
    
    # Initialize an empty string for the table name
    table_name = ""
    
    # Generate a random table name of the specified length
    for i in 1:length
        # Select a random character from the valid characters
        random_char = valid_chars[rand(1:end)]
        
        # Ensure the first character is a letter or underscore
        if i == 1 && !isletter(random_char) && random_char != '_'
            random_char = valid_chars[rand(1:52)]  # Choose a letter or underscore
        end
        
        # Append the character to the table name
        table_name *= string(random_char)
    end

    return table_name
end

"""
    custom_load(conn::MySQL.Connection, table_name::String, df::DataFrame)

Inserts the contents of a DataFrame into an existing SQL table in a MySQL or MariaDB database. This function constructs and executes a parameterized `INSERT INTO` statement for each row in the DataFrame, ensuring that data is inserted in a manner consistent with the table's schema.

# Arguments
- `conn::MySQL.Connection`: An active connection to a MySQL or MariaDB database.
- `table_name::String`: The name of the table into which data from the DataFrame will be inserted.
- `df::DataFrame`: The DataFrame containing the data to be inserted into the table.

# Details
This function dynamically constructs the column names and placeholder parameters based on the DataFrame's schema. It then prepares a single `INSERT INTO` SQL statement and executes it for each row in the DataFrame, passing the row's values as parameters to ensure type safety and protection against SQL injection.

Column names are wrapped in backticks to allow for the use of reserved words or special characters in column names.

# Example
> conn = MySQL.connect("hostname", "username", "password", "database_name")
> df = DataFrame(region=["North", "South"], geo_id=["N001", "S001"], area=[123.4, 567.8])
> custom_load(conn, "geographical_features", df)

This will insert each row of `df` into the `geographical_features` table in the specified database.

# Notes
- The table specified by `table_name` must already exist in the database, and its schema should be compatible with the structure of the DataFrame `df`.
- This function does not handle the creation of the table or manage conflicts with existing data. If the table does not exist or if there are key conflicts, the function will raise an error.
- To prevent SQL injection and ensure data integrity, this function uses parameterized SQL statements.
"""
function custom_load(conn::MySQL.Connection, table_name::String, df::DataFrame)
    # Preprocess DataFrame to ensure data integrity
    preprocess_dataframe!(df)
    
    col_names = join(["`$(name)`" for name in names(df)], ", ")
    placeholders = join(["?" for _ in 1:length(names(df))], ", ")
    insert_stmt = "INSERT INTO $table_name ($col_names) VALUES ($placeholders);"
    prepared_stmt = DBInterface.prepare(conn, insert_stmt)
    
    for row in eachrow(df)
        # Convert row to Tuple, directly handling `missing` (for NULL handling in SQL)
        values = Tuple([x === missing ? missing : x for x in row])
        DBInterface.execute(prepared_stmt, values)
    end
end

"""
    preprocess_dataframe!(df::DataFrame)

Preprocesses a DataFrame `df` in place to ensure its values conform to specific rules before loading into a SQL database. This function performs the following operations on each column of the DataFrame:

- Clamps numeric values to specified minimum and maximum values based on the column's data type, using a predefined `SQL_RANGES` dictionary that maps Julia types to their corresponding SQL range limits.

- Replaces `NaN` and `Inf` values in floating-point columns with `nothing`, making them compatible with SQL `NULL` values. This step ensures that the DataFrame can be safely inserted into a SQL database without encountering errors related to unsupported or undefined numeric values.

# Arguments
- `df::DataFrame`: The DataFrame to be preprocessed. The function modifies `df` in place.

# Examples
Given a DataFrame `df` with possible out-of-range and NaN values:
> df = DataFrame(
    A = [1.0, 2.0, NaN, 4.0],
    B = [Inf, -Inf, 5.0, 6.0]
  )

And a predefined `SQL_RANGES` dictionary specifying the allowed range for each type:
> const SQL_RANGES = Dict{Type, Tuple{Number, Number}}(
    Float64 => (-100.0, 100.0)  # Example range for demonstration
  )

Call `preprocess_dataframe!` to preprocess `df` in place:
> preprocess_dataframe!(df)

After preprocessing, `df` will have its `NaN` and `Inf` values replaced with `nothing`, and numeric values will be clamped to the specified ranges in `SQL_RANGES`.

# Notes
- The `SQL_RANGES` dictionary must be defined and accessible in the scope where `preprocess_dataframe!` is called. It should map Julia types to tuples of `(min_value, max_value)`, specifying the allowed range for each type.

- The function operates in place, modifying the original DataFrame. Ensure you have made any necessary backups of the original data before preprocessing.
"""
function preprocess_dataframe!(df::DataFrame)
    for col in names(df)
        col_type = eltype(df[!, col])

        # Convert missing values to nothing for SQL NULL compatibility
        df[!, col] = map(x -> ismissing(x) ? missing : x, df[!, col])

        # Clamp numeric values if range is defined for the column type
        if haskey(SQL_RANGES, col_type)
            min_val, max_val = SQL_RANGES[col_type]
            df[!, col] = clamp.(df[!, col], min_val, max_val)
        end

        # Replace NaN and Inf with nothing in numeric columns
        if col_type <: AbstractFloat
            df[!, col] = map(x -> isfinite(x) ? x : missing, df[!, col])
        end
    end
end
